/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { registerLocaleData } from '@angular/common';
import localeDe from '@angular/common/locales/de';
import localeDeExtra from '@angular/common/locales/extra/de';
import { VorgangFormularButtonsE2EComponent } from 'apps/alfa-e2e/src/components/vorgang/vorgang-formular-buttons.e2e.components';
import { SnackBarE2EComponent } from '../../../components/ui/snackbar.e2e.component';
import { VorgangListE2EComponent } from '../../../components/vorgang/vorgang-list.e2e.component';
import {
  VorgangE2E,
  VorgangMessagesE2E,
  VorgangStatusE2E,
  vorgangStatusLabelE2E,
} from '../../../model/vorgang';
import { MainPage, waitForSpinnerToDisappear } from '../../../page-objects/main.po';
import { VorgangPage } from '../../../page-objects/vorgang.po';
import { dropCollections } from '../../../support/cypress-helper';
import { contains, exist, haveText, notExist } from '../../../support/cypress.util';
import { loginAsSabine } from '../../../support/user-util';
import {
  buildVorgang,
  createVorgang,
  initVorgaenge,
  objectIds,
} from '../../../support/vorgang-util';

registerLocaleData(localeDe, 'de', localeDeExtra);

describe('Vorgang Verwerfen', () => {
  const mainPage: MainPage = new MainPage();
  const vorgangList: VorgangListE2EComponent = mainPage.getVorgangList();
  const snackbar: SnackBarE2EComponent = mainPage.getSnackBar();

  const vorgangPage: VorgangPage = new VorgangPage();
  const vorgangFormularButtons: VorgangFormularButtonsE2EComponent =
    vorgangPage.getFormularButtons();

  const vorgangVerwerfen: VorgangE2E = { ...createVorgang(), name: 'DoVerwerfen' };
  const vorgangVerwerfenRevoke: VorgangE2E = { ...buildVorgang(objectIds[0], 'DoRevokeVerwerfen') };

  before(() => {
    initVorgaenge([vorgangVerwerfen, vorgangVerwerfenRevoke]);

    loginAsSabine();

    waitForSpinnerToDisappear();
    exist(vorgangList.getRoot());
  });

  after(() => {
    dropCollections();
  });

  describe('verwerfen', () => {
    it('Open Vorgang-Detail-Page', () => {
      vorgangList.getListItem(vorgangVerwerfen.name).getRoot().click();
      waitForSpinnerToDisappear();

      exist(vorgangPage.getVorgangDetailHeader().getRoot());
    });

    describe('by button', () => {
      it('should have status Neu', () => {
        haveText(
          vorgangPage.getVorgangDetailHeader().getStatus(),
          vorgangStatusLabelE2E[vorgangVerwerfen.status],
        );
      });

      it('should show snackBar', () => {
        vorgangFormularButtons.getVerwerfenButton().click();
        waitForSpinnerToDisappear();

        exist(snackbar.getCloseButton());
        contains(snackbar.getMessage(), VorgangMessagesE2E.VERWORFEN);
      });

      it('should close snackBar on close button', () => {
        snackbar.getCloseButton().click();

        notExist(snackbar.getMessage());
      });

      it('should have status Verworfen', () => {
        haveText(
          vorgangPage.getVorgangDetailHeader().getStatus(),
          vorgangStatusLabelE2E[VorgangStatusE2E.VERWORFEN],
        );
      });

      it('should have button zurückholen', () => {
        exist(vorgangPage.getSubnavigation().getZurueckholenIconButton());
        exist(vorgangFormularButtons.getZurueckholenButton());
      });

      it('back to vorgang list', () => {
        vorgangPage.getSubnavigation().getBackButton().click();
        waitForSpinnerToDisappear();

        exist(vorgangList.getRoot());
      });

      it('should have status Verworfen', () => {
        haveText(
          vorgangList.getListItem(vorgangVerwerfen.name).getStatus(),
          vorgangStatusLabelE2E[VorgangStatusE2E.VERWORFEN],
        );
      });
    });
  });

  describe('verwerfen und rückgängig machen', () => {
    it('Open Vorgang-Detail-Page', () => {
      vorgangList.getListItem(vorgangVerwerfenRevoke.name).getRoot().click();
      waitForSpinnerToDisappear();

      exist(vorgangPage.getVorgangDetailHeader().getRoot());
    });

    describe('by icon-button', () => {
      it('should have status Neu', () => {
        haveText(
          vorgangPage.getVorgangDetailHeader().getStatus(),
          vorgangStatusLabelE2E[vorgangVerwerfenRevoke.status],
        );
      });

      it('should show snackBar', () => {
        vorgangPage.getSubnavigation().getVerwerfenIconButton().click();
        waitForSpinnerToDisappear();

        exist(snackbar.getCloseButton());
        contains(snackbar.getMessage(), VorgangMessagesE2E.VERWORFEN);
      });

      it('should close snackbar on revoke', () => {
        snackbar.getRevokeButton().click();
        waitForSpinnerToDisappear();

        notExist(snackbar.getMessage());
      });

      it('should show status Neu', () => {
        notExist(vorgangPage.getProgressBar());

        haveText(
          vorgangPage.getVorgangDetailHeader().getStatus(),
          vorgangStatusLabelE2E[VorgangStatusE2E.NEU],
        );
      });

      it('should not have button zurückholen', () => {
        notExist(vorgangPage.getSubnavigation().getZurueckholenIconButton());
        notExist(vorgangFormularButtons.getZurueckholenButton());
      });

      it('back to vorgang list', () => {
        vorgangPage.getSubnavigation().getBackButton().click();
        waitForSpinnerToDisappear();

        exist(vorgangList.getRoot());
      });

      it('should have status Neu', () => {
        haveText(
          vorgangList.getListItem(vorgangVerwerfenRevoke.name).getStatus(),
          vorgangStatusLabelE2E[VorgangStatusE2E.NEU],
        );
      });
    });
  });
});
