import { mock, Mock, useFromMock } from '@alfa-client/test-utils';
import { fakeAsync, tick } from '@angular/core/testing';
import { cold } from 'jest-marbles';
import { singleCold } from 'libs/tech-shared/test/marbles';
import { firstValueFrom, lastValueFrom, Observable, of } from 'rxjs';
import {
  createOrganisationseinheit,
  createOrganisationseinheitState,
} from '../../../test/user/user';
import { Organisationseinheit } from './user.model';
import { UserRepository } from './user.repository.service';
import { UserService } from './user.service';

describe('UserService', () => {
  let service: UserService;
  let repository: Mock<UserRepository>;
  const sortedNames: string[] = ['BBBB', 'CCCC', 'XXXX'];
  const sortedOrganisationseinheitItems: Organisationseinheit[] = sortedNames.map((name) => ({
    ...createOrganisationseinheit(),
    name,
  }));
  const unsortedOrganisationseinheitItems: Organisationseinheit[] = [
    sortedOrganisationseinheitItems[2],
    sortedOrganisationseinheitItems[0],
    sortedOrganisationseinheitItems[1],
  ];
  let setOrganisationseinheitItemsSpy: jest.SpyInstance;
  let setLoadingUntilFirstSpy: jest.SpyInstance;

  beforeEach(() => {
    repository = mock(UserRepository);
    service = new UserService(useFromMock(repository));
    setOrganisationseinheitItemsSpy = jest.spyOn(service, 'setOrganisationseinheit');
    setLoadingUntilFirstSpy = jest.spyOn(service, 'setLoadingUntilFirst');
  });

  it('should be created', () => {
    expect(service).toBeTruthy();
  });

  describe('initial state', () => {
    it('should have empty group list', () => {
      expect(service.organisationseinheitState$.value.organisationseinheitItems).toEqual([]);
    });
    it('should not be loading', () => {
      expect(service.organisationseinheitState$.value.loading).toEqual(false);
    });
    it('should be requiring a reload', () => {
      expect(service.organisationseinheitState$.value.updateRequired).toEqual(true);
    });
  });

  describe('get group state', () => {
    it('should update if required', async () => {
      service.updateIfRequired = jest.fn();

      await firstValueFrom(service.getOrganisationseinheitState());

      expect(service.updateIfRequired).toHaveBeenCalled();
    });

    it('should call find groups', fakeAsync(() => {
      repository.findOrganisationseinheitItems.mockReturnValue(
        of(unsortedOrganisationseinheitItems),
      );

      service.getOrganisationseinheitState().subscribe();
      tick();

      expect(repository.findOrganisationseinheitItems).toHaveBeenCalled();
    }));
  });

  describe('update if required', () => {
    beforeEach(() => {
      service.updateOrganisationseinheitItems = jest.fn(() => of([]));
    });

    it('should call update groups with initial values', () => {
      service.updateIfRequired();

      expect(service.updateOrganisationseinheitItems).toHaveBeenCalled();
    });

    it('should not call update groups if loading', () => {
      service.setLoading();

      service.updateIfRequired();

      expect(service.updateOrganisationseinheitItems).not.toHaveBeenCalled();
    });

    it('should not call update groups if not requiring update', () => {
      service.setUpdateRequired(false);

      service.updateIfRequired();

      expect(service.updateOrganisationseinheitItems).not.toHaveBeenCalled();
    });
  });

  describe('update groups', () => {
    beforeEach(() => {
      repository.findOrganisationseinheitItems.mockReturnValue(
        of(unsortedOrganisationseinheitItems),
      );
    });

    it('should set loading', () => {
      service.updateOrganisationseinheitItems();

      expect(service.organisationseinheitState$.value.loading).toBeTruthy();
    });

    it('should set sorted groups', async () => {
      await lastValueFrom(service.updateOrganisationseinheitItems());

      expect(service.organisationseinheitState$.value.organisationseinheitItems).toEqual(
        sortedOrganisationseinheitItems,
      );
    });

    it('should not require update afterward', async () => {
      await lastValueFrom(service.updateOrganisationseinheitItems());

      expect(service.organisationseinheitState$.value.updateRequired).toBeFalsy();
    });
  });

  describe('save organisationseinheit', () => {
    const saveGroup: Organisationseinheit = {
      ...createOrganisationseinheit(),
      id: sortedOrganisationseinheitItems[0].id,
    };

    beforeEach(() => {
      repository.saveOrganisationseinheit.mockReturnValue(of(null));
    });

    it('should use set loading until first', fakeAsync(() => {
      service.saveOrganisationseinheit(saveGroup).subscribe();
      tick();

      expect(setLoadingUntilFirstSpy).toHaveBeenCalled();
    }));

    it('should call repository for save', fakeAsync(() => {
      service.saveOrganisationseinheit(saveGroup).subscribe();
      tick();

      expect(repository.saveOrganisationseinheit).toHaveBeenCalledWith(saveGroup);
    }));

    it('should set organisationseinheit items with updated organisationseinheit', fakeAsync(() => {
      service.setOrganisationseinheit(sortedOrganisationseinheitItems);
      setOrganisationseinheitItemsSpy.mockClear();

      service.saveOrganisationseinheit(saveGroup).subscribe();
      tick();

      expect(setOrganisationseinheitItemsSpy).toHaveBeenCalledWith([
        saveGroup,
        ...sortedOrganisationseinheitItems.slice(1),
      ]);
    }));
  });

  describe('create organisationseinheit', () => {
    const organisationseinheit: Organisationseinheit = createOrganisationseinheit();

    beforeEach(() => {
      repository.createOrganisationseinheit.mockReturnValue(of(organisationseinheit));
    });

    it('should use set loading until first', () => {
      service.createOrganisationseinheit(
        organisationseinheit.name,
        organisationseinheit.organisationseinheitIds,
      );

      expect(setLoadingUntilFirstSpy).toHaveBeenCalled();
    });

    it('should call repository for save', () => {
      service.createOrganisationseinheit(
        organisationseinheit.name,
        organisationseinheit.organisationseinheitIds,
      );

      expect(repository.createOrganisationseinheit).toHaveBeenCalledWith(
        organisationseinheit.name,
        organisationseinheit.organisationseinheitIds,
      );
    });

    it('should set organisationseinheit items with new group', fakeAsync(() => {
      service.setOrganisationseinheit(sortedOrganisationseinheitItems);
      setOrganisationseinheitItemsSpy.mockClear();

      service
        .createOrganisationseinheit(
          organisationseinheit.name,
          organisationseinheit.organisationseinheitIds,
        )
        .subscribe();
      tick();

      expect(setOrganisationseinheitItemsSpy).toHaveBeenCalledWith([
        ...sortedOrganisationseinheitItems,
        organisationseinheit,
      ]);
    }));
  });

  describe('set organisationseinheit items', () => {
    it('should sort groups by name', async () => {
      service.setOrganisationseinheit(unsortedOrganisationseinheitItems);

      expect(service.organisationseinheitState$.value.organisationseinheitItems).toEqual(
        sortedOrganisationseinheitItems,
      );
    });
  });

  describe('delete group', () => {
    const deleteGroup: Organisationseinheit = unsortedOrganisationseinheitItems[1];

    beforeEach(async () => {
      repository.deleteOrganisationseinheit.mockReturnValue(of(null));
      service.organisationseinheitState$.next({
        ...service.organisationseinheitState$.value,
        organisationseinheitItems: sortedOrganisationseinheitItems,
      });
    });

    it('should use set loading until first', () => {
      const setLoadingUntilFirstSpy: jest.SpyInstance = jest.spyOn(service, 'setLoadingUntilFirst');
      service.deleteOrganisationseinheit(deleteGroup.id);

      expect(setLoadingUntilFirstSpy).toHaveBeenCalled();
    });

    it('should call repository', () => {
      service.deleteOrganisationseinheit(deleteGroup.id);

      expect(repository.deleteOrganisationseinheit).toHaveBeenCalledWith(deleteGroup.id);
    });

    describe('which exists', () => {
      it('should not have group in list after delete', fakeAsync(() => {
        service.deleteOrganisationseinheit(deleteGroup.id).subscribe();
        tick();

        expect(service.organisationseinheitState$.value.organisationseinheitItems).not.toContain(
          deleteGroup,
        );
      }));
    });

    describe('which does not exist', () => {
      it('should have no effect', fakeAsync(() => {
        service.deleteOrganisationseinheit('unknown-id').subscribe();
        tick();

        expect(service.organisationseinheitState$.value.organisationseinheitItems).toEqual(
          sortedOrganisationseinheitItems,
        );
      }));
    });
  });

  describe('set loading until first', () => {
    it('should set loading', () => {
      service.setLoadingUntilFirst(of(null), () => {});

      expect(service.organisationseinheitState$.value.loading).toBe(true);
    });

    it('should unset loading', fakeAsync(() => {
      service.setLoadingUntilFirst(of(null), () => {}).subscribe();
      tick();

      expect(service.organisationseinheitState$.value.loading).toBe(false);
    }));

    it('should emit loading before first and emit loading after first', () => {
      const delayedNull = cold('-a|', { a: null });

      const progressObservable = service.setLoadingUntilFirst(delayedNull, () => {});

      expect(progressObservable).toBeObservable(cold('a(b|)', { a: true, b: false }));
    });

    it('should call tap function', fakeAsync(() => {
      const tapFunction: jest.Mock = jest.fn();
      const value: string = 'abc';
      const delayedNull: Observable<string> = of(value);

      service.setLoadingUntilFirst(delayedNull, tapFunction).subscribe();
      tick();

      expect(tapFunction).toHaveBeenCalledWith(value);
    }));
  });

  describe('get organisationseinheit items', () => {
    it('should return items of state', () => {
      const organisationseinheitItems: Organisationseinheit[] = [
        createOrganisationseinheit(),
        createOrganisationseinheit(),
      ];
      service.getOrganisationseinheitState = jest
        .fn()
        .mockReturnValue(singleCold(createOrganisationseinheitState(organisationseinheitItems)));

      const itemsObservable: Observable<Organisationseinheit[]> =
        service.getOrganisationseinheitItems();
      expect(itemsObservable).toBeObservable(singleCold(organisationseinheitItems));
    });
  });
});
