/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { BinaryFileResource } from '@alfa-client/binary-file-shared';
import {
  createEmptyStateResource,
  createStateResource,
  StateResource,
} from '@alfa-client/tech-shared';
import { Mock, mock, useFromMock } from '@alfa-client/test-utils';
import { fakeAsync, tick } from '@angular/core/testing';
import { getUrl } from '@ngxp/rest';
import { cold, hot } from 'jest-marbles';
import { Observable, of } from 'rxjs';
import { createBinaryFileResource } from '../../../binary-file-shared/test/binary-file';
import { singleCold } from '../../../tech-shared/test/marbles';
import { createApiRootResource } from '../../test/api-root';
import { ApiRootFacade } from './+state/api-root.facade';
import { ApiRootResource } from './api-root.model';
import { ApiRootRepository } from './api-root.repository';
import { ApiRootService } from './api-root.service';

describe('ApiRootService', () => {
  let service: ApiRootService;
  let repository: Mock<ApiRootRepository>;
  let facade: Mock<ApiRootFacade>;

  const apiRoot: ApiRootResource = createApiRootResource();
  const apiRootStateResource: StateResource<ApiRootResource> = createStateResource(apiRoot);

  beforeEach(() => {
    repository = mock(ApiRootRepository);
    facade = mock(ApiRootFacade);

    service = new ApiRootService(useFromMock(repository), useFromMock(facade));
  });

  describe('get apiRoot', () => {
    beforeEach(() => {
      service.apiRootStateResource.next(apiRootStateResource);
      service.handleApiRootChanges = jest.fn();
      service.shouldEmit = jest.fn().mockReturnValue(true);
    });

    it('should call handle apiRoot changes', fakeAsync(() => {
      service.getApiRoot().subscribe();
      tick();

      expect(service.handleApiRootChanges).toHaveBeenCalledWith(apiRootStateResource);
    }));

    it('should call shouldEmit', fakeAsync(() => {
      service.getApiRoot().subscribe();
      tick();

      expect(service.shouldEmit).toHaveBeenCalledWith(apiRootStateResource);
    }));

    it('should return value', () => {
      service.apiRootStateResource.asObservable = jest
        .fn()
        .mockReturnValue(singleCold(apiRootStateResource, '-a'));

      const apiRootStateResource$: Observable<StateResource<ApiRootResource>> =
        service.getApiRoot();

      expect(apiRootStateResource$).toBeObservable(
        cold('ab', { a: createEmptyStateResource(true), b: apiRootStateResource }),
      );
    });
  });

  describe('handle apiRoot changes', () => {
    it('should load apiRoot if loading required', () => {
      service.loadApiRoot = jest.fn();

      service.handleApiRootChanges(createEmptyStateResource());

      expect(service.loadApiRoot).toHaveBeenCalled();
    });
    it('should NOT load apiRoot if already loaded', () => {
      service.loadApiRoot = jest.fn();

      service.handleApiRootChanges(apiRootStateResource);

      expect(service.loadApiRoot).not.toHaveBeenCalled();
    });
  });

  describe('load apiRoot', () => {
    beforeEach(() => {
      service.apiRootStateResource.next(createEmptyStateResource());
      service.doLoadApiRoot = jest.fn();
    });

    it('should set apiRoot stateResource loading', () => {
      service.loadApiRoot();

      expect(service.apiRootStateResource.value.loading).toBeTruthy();
    });

    it('should do load apiRoot', () => {
      service.loadApiRoot();

      expect(service.doLoadApiRoot).toHaveBeenCalled();
    });
  });

  describe('do load apiRoot', () => {
    const apiRoot: ApiRootResource = createApiRootResource();

    beforeEach(() => {
      service.apiRootStateResource.next(createEmptyStateResource());
      repository.loadApiRoot.mockReturnValue(of(apiRoot));
    });
    it('should call repository', () => {
      service.doLoadApiRoot();

      expect(repository.loadApiRoot).toHaveBeenCalled();
    });

    it('should update apiRoot stateResource', () => {
      service.doLoadApiRoot();

      expect(service.apiRootStateResource.value.resource).toBe(apiRoot);
    });
  });

  describe('should emit', () => {
    it('should return true on loaded stateResource', () => {
      const shouldEmit: boolean = service.shouldEmit(createStateResource(createApiRootResource()));

      expect(shouldEmit).toBeTruthy();
    });

    it('should return true on loading stateResource', () => {
      const shouldEmit: boolean = service.shouldEmit(createEmptyStateResource(true));

      expect(shouldEmit).toBeTruthy();
    });

    it('should return false on reload stateResource', () => {
      const shouldEmit: boolean = service.shouldEmit({
        ...createStateResource(createApiRootResource()),
        reload: true,
      });

      expect(shouldEmit).toBeFalsy();
    });
  });

  describe('getDownloadToken', () => {
    const binaryFile: BinaryFileResource = createBinaryFileResource();

    beforeEach(() => {
      facade.getApiRoot.mockReturnValue(hot('-a', { a: apiRootStateResource }));
    });

    it('should call getApiRoot', () => {
      service.getDownloadToken(binaryFile);

      expect(facade.getApiRoot).toHaveBeenCalled();
    });

    it('should NOT call repository on null apiRoot', () => {
      service.getApiRoot = jest.fn().mockReturnValue(of(createEmptyStateResource()));

      service.getDownloadToken(binaryFile).subscribe();

      expect(repository.downloadToken).not.toHaveBeenCalled();
    });

    it('should call repository', () => {
      facade.getApiRoot.mockReturnValue(of(apiRootStateResource));

      service.getDownloadToken(binaryFile).subscribe();

      expect(repository.downloadToken).toHaveBeenCalledWith(apiRoot, getUrl(binaryFile));
    });
  });
});
