import { CommandOrder, CommandResource, CommandService } from '@alfa-client/command-shared';
import { StateResource, createStateResource } from '@alfa-client/tech-shared';
import { Mock, mock, useFromMock } from '@alfa-client/test-utils';
import { createCommandResource } from 'libs/command-shared/test/command';
import { singleColdCompleted } from 'libs/tech-shared/test/marbles';
import { Observable, of } from 'rxjs';
import {
  createCollaboration,
  createCollaborationListResource,
} from '../../../collaboration-shared/test/collaboration';
import { CollaborationListResourceService } from './collaboration-list-resource.service';
import { CollaborationListLinkRel } from './collaboration.linkrel';
import { Collaboration, CollaborationListResource } from './collaboration.model';
import { CollaborationService } from './collaboration.service';

jest.mock('./collaboration-list-resource.service');

describe('CollaborationService', () => {
  let service: CollaborationService;

  let listService: Mock<CollaborationListResourceService>;
  let commandService: Mock<CommandService>;

  const collaborationListResource: CollaborationListResource = createCollaborationListResource();
  const collaborationStateListResource: StateResource<CollaborationListResource> =
    createStateResource(collaborationListResource);

  beforeEach(() => {
    listService = mock(CollaborationListResourceService);
    commandService = mock(CommandService);

    service = new CollaborationService(useFromMock(listService), useFromMock(commandService));
  });

  describe('get list', () => {
    beforeEach(() => {
      listService.getList.mockReturnValue(of(collaborationStateListResource));
    });

    it('should call service', () => {
      service.getList();

      expect(listService.getList).toHaveBeenCalled();
    });

    it('should return value', () => {
      const list$: Observable<StateResource<CollaborationListResource>> = service.getList();

      expect(list$).toBeObservable(singleColdCompleted(collaborationStateListResource));
    });
  });

  describe('is request form visible', () => {
    it('should return value', (done) => {
      service.showRequestForm$.next(false);

      service.isRequestFormVisible().subscribe((isVisible: boolean) => {
        expect(isVisible).toBeTruthy();
        done();
      });

      service.showRequestForm();
    });
  });

  describe('show anfrage formular', () => {
    it('should set "showRequestForm" to true', () => {
      service.showRequestForm$.next(false);

      service.showRequestForm();

      expect(service.showRequestForm$.value).toBeTruthy();
    });
  });

  describe('hide anfrage formular', () => {
    it('should set "showRequestForm" to false', () => {
      service.showRequestForm$.next(true);

      service.hideRequestForm();

      expect(service.showRequestForm$.value).toBeFalsy();
    });
  });

  describe('create', () => {
    const collaborationListResource: CollaborationListResource = createCollaborationListResource();
    const commandStateResource: StateResource<CommandResource> =
      createStateResource(createCommandResource());
    const collaboration: Collaboration = createCollaboration();

    beforeEach(() => {
      commandService.createCommandByProps.mockReturnValue(of(commandStateResource));
    });

    it('should call command service', () => {
      service.create(collaborationListResource, collaboration);

      expect(commandService.createCommandByProps).toHaveBeenCalledWith({
        linkRel: CollaborationListLinkRel.CREATE_COLLABORATION_REQUEST,
        resource: collaborationListResource,
        command: {
          order: CommandOrder.CREATE_COLLABORATION_REQUEST,
          body: collaboration,
        },
        snackBarMessage: 'Die Zuarbeit wurde angefragt.',
      });
    });

    it('should return value', () => {
      const created$: Observable<StateResource<CommandResource>> = service.create(
        collaborationListResource,
        collaboration,
      );

      expect(created$).toBeObservable(singleColdCompleted(commandStateResource));
    });
  });
});
