import { VerticalBinaryFileListComponent } from '@alfa-client/binary-file';
import { BinaryFileListResource } from '@alfa-client/binary-file-shared';
import { StateResource, createStateResource } from '@alfa-client/tech-shared';
import { Mock, getDebugElementFromFixtureByCss, mock } from '@alfa-client/test-utils';
import {
  VorgangService,
  VorgangWithEingangLinkRel,
  VorgangWithEingangResource,
} from '@alfa-client/vorgang-shared';
import { ComponentFixture, TestBed } from '@angular/core/testing';
import faker from '@faker-js/faker';
import { getUrl } from '@ngxp/rest';
import { createBinaryFileListResource } from 'libs/binary-file-shared/test/binary-file';
import { getDataTestIdOf } from 'libs/tech-shared/test/data-test';
import { createVorgangWithEingangResource } from 'libs/vorgang-shared/test/vorgang';
import { MockComponent } from 'ng-mocks';
import { of } from 'rxjs';
import { VorgangDetailDateienContainerComponent } from './vorgang-detail-dateien-container.component';

describe('VorgangDetailDateienContainerComponent', () => {
  let component: VorgangDetailDateienContainerComponent;
  let fixture: ComponentFixture<VorgangDetailDateienContainerComponent>;

  const attachmentList: string = getDataTestIdOf('attachment-list');
  const representationList: string = getDataTestIdOf('representation-list');

  const vorgang: VorgangWithEingangResource = createVorgangWithEingangResource();

  const vorgangService: Mock<VorgangService> = mock(VorgangService);

  beforeEach(async () => {
    await TestBed.configureTestingModule({
      imports: [VorgangDetailDateienContainerComponent],
      declarations: [
        VorgangDetailDateienContainerComponent,
        MockComponent(VerticalBinaryFileListComponent),
      ],
      providers: [
        {
          provide: VorgangService,
          useValue: vorgangService,
        },
      ],
    }).compileComponents();

    fixture = TestBed.createComponent(VorgangDetailDateienContainerComponent);
    component = fixture.componentInstance;
    fixture.detectChanges();
  });

  it('should create', () => {
    expect(component).toBeTruthy();
  });

  it('should call handleVorgangChange on set new vorgang', () => {
    component.handleVorgangChange = jest.fn();

    component.vorgangWithEingang = createVorgangWithEingangResource();

    expect(component.handleVorgangChange).toHaveBeenCalled();
  });

  describe('handle vorgang change', () => {
    it('should set downloadPrefix', () => {
      component.handleVorgangChange(vorgang);

      expect(component.downloadPrefix).toEqual(vorgang.nummer);
    });

    it('should call handle attachements', () => {
      component.handleAttachments = jest.fn();

      component.handleVorgangChange(vorgang);

      expect(component.handleAttachments).toHaveBeenCalledWith(vorgang);
    });

    it('should call get representations if exists', () => {
      component.getRepresentationsIfExists = jest.fn();

      component.handleVorgangChange(vorgang);

      expect(component.getRepresentationsIfExists).toHaveBeenCalledWith(vorgang);
    });
  });

  describe('get representations if exists', () => {
    it('should call vorgangService if link exists', () => {
      component.getRepresentationsIfExists(
        createVorgangWithEingangResource([VorgangWithEingangLinkRel.REPRESENTATIONS]),
      );

      expect(vorgangService.getRepresentations).toHaveBeenCalled();
    });
  });

  describe('handle attachments', () => {
    it('should call set download uri', () => {
      component.setAttachmentsDownloadUri = jest.fn();

      component.handleAttachments(vorgang);

      expect(component.setAttachmentsDownloadUri).toHaveBeenCalledWith(vorgang);
    });

    it('should call vorgangService methods', () => {
      component.handleVorgangChange(
        createVorgangWithEingangResource([VorgangWithEingangLinkRel.ATTACHMENTS]),
      );

      expect(vorgangService.getAttachments).toHaveBeenCalled();
    });

    it('should call get attachments if exists', () => {
      component.getAttachmentsIfExists = jest.fn();

      component.handleVorgangChange(vorgang);

      expect(component.getAttachmentsIfExists).toHaveBeenCalledWith(vorgang);
    });
  });

  describe('get attachments if exists', () => {
    it('should call vorgangService if link exists', () => {
      component.getAttachmentsIfExists(
        createVorgangWithEingangResource([VorgangWithEingangLinkRel.ATTACHMENTS]),
      );

      expect(vorgangService.getAttachments).toHaveBeenCalled();
    });
  });

  describe('download attachments link', () => {
    it('should be set if exists', () => {
      const vorganWithLink: VorgangWithEingangResource = createVorgangWithEingangResource([
        VorgangWithEingangLinkRel.DOWNLOAD_ATTACHMENTS,
      ]);

      component.handleVorgangChange(vorganWithLink);

      expect(component.attachmentsDownloadUri).toBe(
        getUrl(vorganWithLink, VorgangWithEingangLinkRel.DOWNLOAD_ATTACHMENTS),
      );
    });

    it('should NOT be set if missing', () => {
      const vorganWithLink: VorgangWithEingangResource = createVorgangWithEingangResource();

      component.handleVorgangChange(vorganWithLink);

      expect(component.attachmentsDownloadUri).toBeUndefined();
    });
  });

  describe('on existing attachments', () => {
    describe('binary file list component should be called with', () => {
      const binaryFileListResource: BinaryFileListResource = createBinaryFileListResource();
      const binaryFileListStateResource: StateResource<BinaryFileListResource> =
        createStateResource(binaryFileListResource);

      beforeEach(() => {
        component.attachmentListStateResource$ = of(binaryFileListStateResource);
        fixture.detectChanges();
      });

      it('binaryFileListStateResource', () => {
        const binaryFileListComp: VerticalBinaryFileListComponent =
          getVerticalBinaryFileListComponent(attachmentList);
        expect(binaryFileListComp.binaryFileListStateResource).toEqual(binaryFileListStateResource);
      });

      it('downloadFileNamePrefix', () => {
        const binaryFileListComp: VerticalBinaryFileListComponent =
          getVerticalBinaryFileListComponent(attachmentList);
        expect(binaryFileListComp.downloadFileNamePrefix).toEqual(component.downloadPrefix);
      });

      it('archiveDownloadUri', () => {
        component.attachmentsDownloadUri = faker.internet.url();

        fixture.detectChanges();

        const binaryFileListComp: VerticalBinaryFileListComponent =
          getVerticalBinaryFileListComponent(attachmentList);
        expect(binaryFileListComp.archiveDownloadUri).toEqual(component.attachmentsDownloadUri);
      });
    });
  });

  describe('On existing representations', () => {
    describe('binary file list component should be called with', () => {
      const binaryFileListResource: BinaryFileListResource = createBinaryFileListResource();
      const binaryFileListStateResource: StateResource<BinaryFileListResource> =
        createStateResource(binaryFileListResource);

      beforeEach(() => {
        component.representationListStateResource$ = of(binaryFileListStateResource);
        fixture.detectChanges();
      });

      it('binaryFileListStateResource', () => {
        const binaryFileListComp: VerticalBinaryFileListComponent =
          getVerticalBinaryFileListComponent(representationList);
        expect(binaryFileListComp.binaryFileListStateResource).toEqual(binaryFileListStateResource);
      });

      it('downloadFileNamePrefix', () => {
        const binaryFileListComp: VerticalBinaryFileListComponent =
          getVerticalBinaryFileListComponent(representationList);
        expect(binaryFileListComp.downloadFileNamePrefix).toEqual(component.downloadPrefix);
      });
    });
  });

  function getVerticalBinaryFileListComponent(
    dataTestIdElement: string,
  ): VerticalBinaryFileListComponent {
    return getDebugElementFromFixtureByCss(
      fixture,
      'alfa-vertical-binary-file-list' + dataTestIdElement,
    ).componentInstance;
  }
});
