/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.stereotype.Component;

import com.auth0.jwt.JWT;
import com.auth0.jwt.algorithms.Algorithm;
import com.auth0.jwt.exceptions.JWTVerificationException;

import de.ozgcloud.alfa.common.binaryfile.FileId;
import de.ozgcloud.alfa.common.downloadtoken.DownloadTokenProperties;
import de.ozgcloud.alfa.common.user.UserProfile;
import io.jsonwebtoken.Claims;
import io.jsonwebtoken.Jwts;
import io.jsonwebtoken.SignatureAlgorithm;

@Component
public class JwtTokenUtil {

	public static final String TOKEN_TYP_KEY = "typ";
	public static final String TOKEN_TYPE = "JWT";
	public static final String TOKEN_ISSUER = "secure-api";
	public static final String TOKEN_AUDIENCE = "secure-app";
	public static final String ROLE_CLAIM = "roles";
	public static final String USERID_CLAIM = "userId";
	public static final String FIRSTNAME_CLAIM = "firstName";
	public static final String LASTNAME_CLAIM = "lastName";
	public static final String FILEID_CLAIM = "fileId";
	public static final String ORGANSIATIONSEINHEIT_IDS_CLAIM = "organisationseinheitIds";

	@Autowired
	private DownloadTokenProperties downloadTokenProperties;

	@SuppressWarnings("unchecked")
	private List<Map<String, String>> getRoleClaims(String token) {
		List<Map<String, String>> roleClaims = new ArrayList<>();

		getAllClaimsFromToken(token).ifPresent(claims -> roleClaims.addAll((List<Map<String, String>>) claims.get(ROLE_CLAIM)));

		return roleClaims;
	}

	public Optional<Claims> getAllClaimsFromToken(String token) {
		return Optional.of(Jwts.parser().setSigningKey(downloadTokenProperties.getSecret().getBytes()).parseClaimsJws(token).getBody());
	}

	public List<SimpleGrantedAuthority> getRolesFromToken(String token) {
		List<Map<String, String>> claimsList = getRoleClaims(token);
		return claimsList.stream()
				.flatMap(rolesMap -> rolesMap.entrySet().stream())
				.map(roleEntry -> new SimpleGrantedAuthority(roleEntry.getValue()))
				.toList();
	}

	@SuppressWarnings({ "unchecked" })
	public Collection<String> getOrganisationseinheitIdsFromToken(String token) {
		List<String> organisationseinheitIds = new ArrayList<>();
		getAllClaimsFromToken(token)
				.ifPresent(claims -> {
					if (claims.get(ORGANSIATIONSEINHEIT_IDS_CLAIM) != null) {
						organisationseinheitIds.addAll((Collection<String>) claims.get(ORGANSIATIONSEINHEIT_IDS_CLAIM));
					}
				});
		return organisationseinheitIds;
	}

	public String generateToken(FileId fileId, UserProfile user) {
		var claims = new HashMap<String, Object>();
		claims.put(USERID_CLAIM, user.getId().toString());
		claims.put(FIRSTNAME_CLAIM, user.getFirstName());
		claims.put(LASTNAME_CLAIM, user.getLastName());
		claims.put(ROLE_CLAIM, user.getAuthorities());
		claims.put(FILEID_CLAIM, fileId.toString());
		claims.put(ORGANSIATIONSEINHEIT_IDS_CLAIM, user.getOrganisationseinheitIds());

		return doGenerateToken(claims, user.getId().toString());
	}

	private String doGenerateToken(Map<String, Object> claims, String subject) {
		return Jwts.builder()
				.setClaims(claims)
				.setSubject(subject)
				.setHeaderParam(TOKEN_TYP_KEY, TOKEN_TYPE)
				.setIssuer(TOKEN_ISSUER).setIssuedAt(new Date(System.currentTimeMillis()))
				.setExpiration(new Date(System.currentTimeMillis() + downloadTokenProperties.getValidity()))
				.setAudience(TOKEN_AUDIENCE)
				.signWith(SignatureAlgorithm.HS512, downloadTokenProperties.getSecret().getBytes())
				.compact();
	}

	public void verifyToken(String token) throws JWTVerificationException {
		var algorithm = Algorithm.HMAC512(downloadTokenProperties.getSecret());
		var verifier = JWT.require(algorithm).build();

		verifier.verify(token);
	}
}