package de.ozgcloud.alfa.bescheid;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.*;

import java.util.Objects;
import java.util.function.BooleanSupplier;

import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.LinkRelation;
import org.springframework.hateoas.server.RepresentationModelProcessor;
import org.springframework.hateoas.server.mvc.WebMvcLinkBuilder;
import org.springframework.stereotype.Component;

import de.ozgcloud.alfa.bescheid.BescheidController.BescheidByVorgangController;
import de.ozgcloud.alfa.common.ModelBuilder;
import de.ozgcloud.alfa.common.command.CommandController.CommandByRelationController;
import de.ozgcloud.alfa.common.user.CurrentUserService;
import de.ozgcloud.alfa.common.user.UserRole;
import de.ozgcloud.alfa.vorgang.Vorgang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import lombok.RequiredArgsConstructor;

@Component
@RequiredArgsConstructor
class BescheidVorgangProcessor implements RepresentationModelProcessor<EntityModel<VorgangWithEingang>> {

	static final LinkRelation REL_DRAFT = LinkRelation.of("bescheidDraft");
	static final LinkRelation REL_CREATE_DRAFT = LinkRelation.of("createBescheidDraft");
	static final LinkRelation REL_BESCHEIDE = LinkRelation.of("bescheide");

	private final BescheidService bescheidService;
	private final CurrentUserService currentUserService;

	@Override
	public EntityModel<VorgangWithEingang> process(EntityModel<VorgangWithEingang> model) {
		var vorgang = model.getContent();
		if (Objects.isNull(vorgang) || currentUserService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER)) {
			return model;
		}
		return ModelBuilder.fromModel(model)
				.ifMatch(isRetrievingDraftAllowed(vorgang))
				.addLink(getDraftLink(vorgang.getId()).withRel(REL_DRAFT))
				.ifMatch(isCreatingDraftAllowed(vorgang))
				.addLink(getCreateBescheidDraftLink(vorgang.getId(), vorgang.getVersion()).withRel(REL_CREATE_DRAFT))
				.ifMatch(() -> existsBescheid(vorgang.getId()))
				.addLink(getBescheideLink(vorgang.getId()).withRel(REL_BESCHEIDE))
				.buildModel();
	}

	private WebMvcLinkBuilder getDraftLink(String vorgangId) {
		return linkTo(methodOn(BescheidController.class).getDraft(vorgangId));
	}

	private WebMvcLinkBuilder getCreateBescheidDraftLink(String vorgangId, long vorgangVersion) {
		return linkTo(methodOn(CommandByRelationController.class).createCommand(vorgangId, vorgangId, vorgangVersion, null));
	}

	boolean existsBescheid(String vorgangId) {
		return bescheidService.existsBescheid(vorgangId);
	}

	private WebMvcLinkBuilder getBescheideLink(String vorgangId) {
		return linkTo(methodOn(BescheidByVorgangController.class).getAll(vorgangId));
	}

	BooleanSupplier isRetrievingDraftAllowed(Vorgang vorgang) {
		return () -> isVorgangInBearbeitung(vorgang) && draftExists(vorgang);
	}

	BooleanSupplier isCreatingDraftAllowed(Vorgang vorgang) {
		return () -> isVorgangInBearbeitung(vorgang) && !draftExists(vorgang);
	}

	boolean isVorgangInBearbeitung(Vorgang vorgang) {
		return Vorgang.VorgangStatus.IN_BEARBEITUNG.equals(vorgang.getStatus());
	}

	boolean draftExists(Vorgang vorgang) {
		return bescheidService.getBescheidDraft(vorgang.getId()).isPresent();
	}
}
