package de.ozgcloud.alfa.collaboration;

import static de.ozgcloud.alfa.common.UserProfileUrlProviderTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;
import org.springframework.web.util.UriComponentsBuilder;

import de.ozgcloud.alfa.common.EntityModelTestFactory;
import de.ozgcloud.alfa.common.UserProfileUrlProvider;
import de.ozgcloud.alfa.common.user.CurrentUserService;
import de.ozgcloud.alfa.common.user.UserRole;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;

class CollaborationVorgangProcessorTest {

	@Spy
	@InjectMocks
	private CollaborationVorgangProcessor processor;

	@Mock
	private CollaborationService collaborationService;

	@Mock
	private CurrentUserService currentUserService;

	private final UserProfileUrlProvider urlProvider = new UserProfileUrlProvider();

	@Nested
	class TestProcess {

		@Nested
		class OnNullVorgang {

			@Test
			void shouldNotAddLinksIfVorgangIsNull() {
				var model = processor.process(EntityModelTestFactory.NULLABLE);

				assertThat(model.hasLinks()).isFalse();
			}
		}

		@Nested
		class OnNotVerwaltungUserRole {

			@BeforeEach
			void mockUserService() {
				when(currentUserService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(false);
			}

			@Test
			void shouldNotAddLinks() {
				var model = processor.process(EntityModel.of(VorgangWithEingangTestFactory.create()));

				assertThat(model.hasLinks()).isFalse();
			}

		}

		@Nested
		class OnNonNullVorgangAndOnVerwaltungUserRole {

			@BeforeEach
			void prepareBuilder() {
				initUserProfileUrlProvider(urlProvider);
			}

			@BeforeEach
			void mockUserService() {
				when(currentUserService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(true);
			}

			@Nested
			class OnCollaborationPresent {

				@BeforeEach
				void setUpMock() {
					when(collaborationService.hasCollaboration(VorgangHeaderTestFactory.ID)).thenReturn(true);
				}

				@Test
				void shouldHaveTwoLinks() {
					var model = callProcessor();

					assertThat(model.getLinks()).hasSize(2);
				}

				@Test
				void shouldNotAddSearchOrganisationsEinheitLink() {
					var model = callProcessor();

					assertThat(model.getLink(CollaborationVorgangProcessor.REL_SEARCH_ORGANISATIONS_EINHEIT)).isEmpty();
				}

				@Test
				void shouldAddCollaborationsLink() {
					var model = callProcessor();

					assertThat(model.getLink(CollaborationVorgangProcessor.REL_COLLABORATIONS)).get()
							.extracting(Link::getHref)
							.isEqualTo(CollaborationController.PATH + "/" + VorgangHeaderTestFactory.ID + "/collaborations");
				}
			}

			@Nested
			class OnCollaborationsNotPresent {

				@BeforeEach
				void setUpMock() {
					when(collaborationService.hasCollaboration(VorgangHeaderTestFactory.ID)).thenReturn(false);
				}

				@Test
				void shouldHaveThreeLinks() {
					var model = callProcessor();

					assertThat(model.getLinks()).hasSize(3);

				}

				@Test
				void shouldAddSearchOrganisationsEinheitLink() {
					var expectedHref = UriComponentsBuilder.fromUriString(OrganisationsEinheitController.PATH)
							.queryParam(OrganisationsEinheitController.SEARCH_BY_PARAM, "{" + OrganisationsEinheitController.SEARCH_BY_PARAM + "}")
							.build().toString();

					var model = callProcessor();

					assertThat(model.getLink(CollaborationVorgangProcessor.REL_SEARCH_ORGANISATIONS_EINHEIT)).get()
							.extracting(Link::getHref)
							.isEqualTo(expectedHref);
				}

				@Test
				void shouldAddCollaborationsLink() {
					var expectedHref = UriComponentsBuilder.fromUriString(CollaborationController.PATH)
							.pathSegment(VorgangHeaderTestFactory.ID, "collaborations")
							.build().toString();

					var model = callProcessor();

					assertThat(model.getLink(CollaborationVorgangProcessor.REL_COLLABORATIONS)).get()
							.extracting(Link::getHref)
							.isEqualTo(expectedHref);
				}
			}

			private EntityModel<VorgangWithEingang> callProcessor() {
				return processor.process(EntityModel.of(VorgangWithEingangTestFactory.create()));
			}
		}
	}
}
