/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.common.command;

import static org.assertj.core.api.Assertions.*;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Spy;

import de.ozgcloud.alfa.common.binaryfile.BinaryFileTestFactory;
import de.ozgcloud.alfa.postfach.PostfachAddress;
import de.ozgcloud.alfa.postfach.PostfachAddressTestFactory;
import de.ozgcloud.alfa.postfach.PostfachMail;
import de.ozgcloud.alfa.postfach.PostfachMailTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.wiedervorlage.WiedervorlageTestFactory;
import de.ozgcloud.vorgang.grpc.command.GrpcCommandBody;
import de.ozgcloud.vorgang.grpc.command.GrpcCommandBodyField;

class CommandBodyMapperTest {

	private static final String FIELD = "FIELD";
	private static final String VALUE = "VALUE";

	@Spy
	private CommandBodyMapper mapper = Mappers.getMapper(CommandBodyMapper.class);

	@DisplayName("From object to map")
	@Nested
	class TestFromObjectToMap {

		@Test
		void shouldMapAllValues() {
			var commandAsObject = CommandTestFactory.create();

			var mappedMap = mapper.fromObjectToMap(commandAsObject);

			assertThat(mappedMap)
					.containsEntry("id", CommandTestFactory.ID)
					.containsEntry("status", CommandTestFactory.STATUS.name())
					.containsEntry("order", CommandTestFactory.ORDER)
					.containsEntry("relationId", CommandTestFactory.RELATION_ID)
					.containsEntry("vorgangId", CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldNotMapClass() {
			var commandAsObject = CommandTestFactory.create();

			var mappedMap = mapper.fromObjectToMap(commandAsObject);

			assertThat(mappedMap).doesNotContainKey("class");
		}

		@Test
		void shouldMapNullValue() {
			var commandAsObject = CommandTestFactory.createBuilder().status(null).build();

			var mappedMap = mapper.fromObjectToMap(commandAsObject);

			assertThat(mappedMap).containsEntry("status", null);
		}

		@Test
		void shouldNotMapVersion() {
			var commandAsObject = WiedervorlageTestFactory.create();

			var mappedMap = mapper.fromObjectToMap(commandAsObject);

			assertThat(mappedMap).doesNotContainKey("version");
		}

		@DisplayName("get value")
		@Nested
		class TestGetValue {

			@Test
			void shouldMapPostfachMailMap() {
				var postfachMailBody = PostfachMailTestFactory.create();

				var mappedValue = mapper.getValue(postfachMailBody);

				assertThat(mappedValue).usingRecursiveComparison().isEqualTo(PostfachMailTestFactory.create());
			}
		}
	}

	@Nested
	class TestMapToBody {
		@Test
		void shouldReturnEmptyBodyForNull() {
			var result = mapper.mapToBody(null);

			assertThat(result).isNotNull();
			assertThat(result.getFieldList()).isEmpty();
		}

		@Test
		void shouldReturnBodyWithField() {
			var result = mapper.mapToBody(buildBodyMap());

			assertThat(result.getFieldList()).hasSize(1);
			assertThat(result.getFieldList().get(0).getName()).isEqualTo("FIELD");
			assertThat(result.getFieldList().get(0).getValue()).isEqualTo("VALUE");
		}

		private Map<String, String> buildBodyMap() {
			return Map.of(FIELD, VALUE);
		}

		@Test
		void shouldFilterEntriesWithNullValue() {
			var result = mapper.mapToBody(buildBodyMapWithNullValue());

			assertThat(result.getFieldList()).isEmpty();
		}

		private Map<String, String> buildBodyMapWithNullValue() {
			HashMap<String, String> bodyMap = new HashMap<>();
			bodyMap.put(FIELD, null);
			return bodyMap;
		}
	}

	@Nested
	class TestMapToBodyFields {

		@Captor
		private ArgumentCaptor<Map<Object, Object>> mapCaptor;

		@Test
		void shouldReturnEmptyListOnNull() {
			var result = mapper.mapToBodyFields((CommandBody) null);

			assertThat(result).isEmpty();
		}

		@Test
		void shouldContainFields() {
			var fieldList = mapper.mapToBodyFields(PostfachMailTestFactory.create());

			assertThat(fieldList).hasSize(13);
		}

		@Test
		void shouldNOTContainClassEntry() {
			var fieldList = mapper.mapToBodyFields(PostfachMailTestFactory.create());

			assertThat(fieldList).isNotEmpty().allMatch(field -> !StringUtils.equals(field.getName(), "class"));
		}

		@Test
		void shouldReturnField() {
			var result = mapper.mapToBodyFields(buildBodyMap());

			assertThat(result).hasSize(1);
			assertThat(result.get(0).getName()).isEqualTo(FIELD);
			assertThat(result.get(0).getValue()).isEqualTo(VALUE);
		}

		private Map<Object, Object> buildBodyMap() {
			return Map.of(FIELD, VALUE);
		}

		@Test
		void shouldMapNullAsEmptyValue() {
			var result = mapper.mapToBodyFields(buildBodyMapWithNullValue());

			assertThat(result).hasSize(1);
			assertThat(result.get(0).getValue()).isEmpty();
		}

		private Map<Object, Object> buildBodyMapWithNullValue() {
			HashMap<Object, Object> bodyMap = new HashMap<>();
			bodyMap.put(FIELD, null);
			return bodyMap;
		}
	}

	@Nested
	class TestMap {

		@Test
		void shouldReturnMap() {
			var result = mapper
					.map(GrpcCommandBody.newBuilder().addField(GrpcCommandBodyField.newBuilder().setName(FIELD).setValue(VALUE).build()).build());

			assertThat(result).hasSize(1).containsEntry(FIELD, VALUE);
		}
	}

	@DisplayName("map to body map")
	@Nested
	class TestMapToBodyMap {

		@DisplayName("for postfachMail")
		@Nested
		class TestForPostfachMail {

			private final CommandBody postfachMailCommandBody = PostfachMailTestFactory.create();
			private final CreateCommand command = CommandTestFactory.createCreateCommandBuilder().body(postfachMailCommandBody).build();
			private final String itemName = "PostfachMail";

			@SuppressWarnings("unchecked")
			@Test
			void shouldMapFields() {
				var bodyMap = mapToBodyMap();

				var item = (Map<String, Object>) bodyMap.get(CommandBodyMapper.ITEM_PROPERTY);
				assertThat(item).containsKey(PostfachMail.POSTFACH_ADDRESS_FIELD);
				var postfachAddressMap = (Map<String, Object>) item.get(PostfachMail.POSTFACH_ADDRESS_FIELD);
				assertThat((Integer) postfachAddressMap.get(PostfachAddress.TYPE_FIELD)).isEqualTo(PostfachAddressTestFactory.TYPE);
				assertThat((String) postfachAddressMap.get(PostfachAddress.VERSION_FIELD)).isEqualTo(PostfachAddressTestFactory.VERSION);
			}

			private Map<String, Object> mapToBodyMap() {
				return mapper.mapToBodyMap(command, itemName);
			}
		}

		@DisplayName("for wiedervorlage")
		@Nested
		class TestForWiedervorlage {

			private final CommandBody wiedervorlageCommandBody = WiedervorlageTestFactory.create();
			private final CreateCommand command = CommandTestFactory.createCreateCommandBuilder().body(wiedervorlageCommandBody).build();
			private final String itemName = "Wiedervorlage";

			@Test
			void shouldContainsItemName() {
				var mappedBody = mapToBodyMap();

				var itemNameEntry = Map.entry(CommandBodyMapper.ITEM_NAME_PROPERTY, itemName);
				assertThat(mappedBody).contains(itemNameEntry);
			}

			@Test
			void shouldContainsVorgangId() {
				var mappedBody = mapToBodyMap();

				var vorgangIdEntry = Map.entry(CommandBodyMapper.VORGANG_ID_PROPERTY, VorgangHeaderTestFactory.ID);
				assertThat(mappedBody).contains(vorgangIdEntry);
			}

			@Nested
			class CreatedItem {

				@Test
				void shouldContainsBetreff() {
					var mappedItemMap = getMappedItemEntry();

					assertThat(mappedItemMap).contains(Map.entry("betreff", WiedervorlageTestFactory.BETREFF));
				}

				@Test
				void shouldContainsBeschreibung() {
					var mappedItemMap = getMappedItemEntry();

					assertThat(mappedItemMap).contains(Map.entry("beschreibung", WiedervorlageTestFactory.BESCHREIBUNG));
				}

				@Test
				void shouldContainsFrist() {
					var mappedItemMap = getMappedItemEntry();

					assertThat(mappedItemMap).contains(Map.entry("frist", WiedervorlageTestFactory.FRIST));
				}

				@Test
				void shouldContainsCreatedBy() {
					var mappedItemMap = getMappedItemEntry();

					assertThat(mappedItemMap).contains(Map.entry("createdBy", WiedervorlageTestFactory.CREATED_BY));
				}

				@Test
				void shouldContainsCreatedAt() {
					var mappedItemMap = getMappedItemEntry();

					assertThat(mappedItemMap).contains(Map.entry("createdAt", WiedervorlageTestFactory.CREATED_AT));
				}

				@Test
				void shouldContainsDone() {
					var mappedItemMap = getMappedItemEntry();

					assertThat(mappedItemMap).contains(Map.entry("done", WiedervorlageTestFactory.DONE));
				}

				@Test
				void shouldContainsAttachments() {
					var mappedItemMap = getMappedItemEntry();

					assertThat(mappedItemMap).contains(Map.entry("attachments", List.of(BinaryFileTestFactory.FILE_ID)));
				}

				@SuppressWarnings("unchecked")
				private Map<String, Object> getMappedItemEntry() {
					return (Map<String, Object>) mapToBodyMap().get(CommandBodyMapper.ITEM_PROPERTY);
				}
			}

			private Map<String, Object> mapToBodyMap() {
				return mapper.mapToBodyMap(command, itemName);
			}
		}
	}
}