/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.common.errorhandling;

import static org.assertj.core.api.Assertions.*;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;

import de.ozgcloud.alfa.common.GrpcUtil;
import io.grpc.Metadata;
import io.grpc.Status;
import io.grpc.StatusRuntimeException;

class GrpcExceptionControllerTest {

	@InjectMocks
	private GrpcExceptionController grpcExceptionController;

	@Nested
	class TestNotFoundStatusException {

		private final StatusRuntimeException notFoundException = GrpcExceptionTestFactory.createGrpcNotFoundStatusRuntimeException();

		@Test
		void shouldHaveMessageCode() {
			var response = grpcExceptionController.handleStatusRuntimeException(notFoundException);

			assertThat(response.getBody().getIssues()).hasSize(1);
			assertThat(response.getBody().getIssues().get(0).getMessageCode())
					.isEqualTo(GrpcExceptionController.GRPC_NOT_FOUND_ERROR_CODE_KEY.getErrorCode());
		}

		@Test
		void shouldHaveMessage() {
			var response = grpcExceptionController.handleStatusRuntimeException(notFoundException);

			assertThat(response.getBody().getIssues()).hasSize(1);
			assertThat(response.getBody().getIssues().get(0).getMessage()).isEqualTo("NOT_FOUND: " + GrpcExceptionTestFactory.NOT_FOUND_DESCRIPTION);
		}

		@Test
		void shouldHaveHttpStatusCode() {
			var response = grpcExceptionController.handleStatusRuntimeException(notFoundException);

			assertThat(response.getStatusCode()).isEqualTo(HttpStatus.NOT_FOUND);
		}

		@Test
		void shouldNotHaveExceptionId() {
			var response = grpcExceptionController.handleStatusRuntimeException(notFoundException);

			assertThat(response.getBody().getIssues()).hasSize(1);
			assertThat(response.getBody().getIssues().get(0).getExceptionId()).isNull();
		}

		@Nested
		class TestWithMetaData {

			private final StatusRuntimeException notFoundExceptionWithMetadata = GrpcExceptionTestFactory
					.createGrpcNotFoundStatusException(GrpcExceptionTestFactory.createMetaData());

			@Test
			void shouldHaveExceptionId() {
				var response = grpcExceptionController.handleStatusRuntimeException(notFoundExceptionWithMetadata);

				assertThat(response.getBody().getIssues()).hasSize(1);
				assertThat(response.getBody().getIssues().get(0).getExceptionId()).isEqualTo(GrpcExceptionTestFactory.METADATA_EXCEPTION_ID_VALUE);
			}
		}
	}

	@Nested
	class TestInternalStatusException {

		private final StatusRuntimeException internalException = GrpcExceptionTestFactory.createGrpcInternalStatusException();

		@Test
		void shouldHaveMessageCode() {
			var response = grpcExceptionController.handleStatusRuntimeException(internalException);

			assertThat(response.getBody().getIssues()).hasSize(1);
			assertThat(response.getBody().getIssues().get(0).getMessageCode())
					.isEqualTo(GrpcExceptionController.GRPC_INTERNAL_ERROR_CODE_KEY.getErrorCode());
		}

		@Test
		void shouldHaveMessage() {
			var response = grpcExceptionController.handleStatusRuntimeException(internalException);

			assertThat(response.getBody().getIssues()).hasSize(1);
			assertThat(response.getBody().getIssues().get(0).getMessage()).isEqualTo("INTERNAL: " + GrpcExceptionTestFactory.TECHNICAL_DESCRIPTION);
		}

		@Test
		void shouldHaveHttpStatusCode() {
			var response = grpcExceptionController.handleStatusRuntimeException(internalException);

			assertThat(response.getStatusCode()).isEqualTo(HttpStatus.INTERNAL_SERVER_ERROR);
		}

		@Test
		void shouldNotHaveExceptionId() {
			var response = grpcExceptionController.handleStatusRuntimeException(internalException);

			assertThat(response.getBody().getIssues()).hasSize(1);
			assertThat(response.getBody().getIssues().get(0).getExceptionId()).isNull();
		}

		@Nested
		class TestWithMetaData {

			private final StatusRuntimeException internalExceptionWithMetadata = GrpcExceptionTestFactory
					.createGrpcInternalStatusException(GrpcExceptionTestFactory.createMetaData());

			@Test
			void shouldHaveErrorCodeAsMessageCode() {
				var response = grpcExceptionController.handleStatusRuntimeException(internalExceptionWithMetadata);

				assertThat(response.getBody().getIssues()).hasSize(1);
				assertThat(response.getBody().getIssues().get(0).getMessageCode()).isEqualTo(GrpcExceptionTestFactory.METADATA_ERROR_CODE_VALUE);
			}

			@Test
			void shouldHaveExceptionId() {
				var response = grpcExceptionController.handleStatusRuntimeException(internalExceptionWithMetadata);

				assertThat(response.getBody().getIssues()).hasSize(1);
				assertThat(response.getBody().getIssues().get(0).getExceptionId()).isEqualTo(GrpcExceptionTestFactory.METADATA_EXCEPTION_ID_VALUE);
			}
		}
	}

	@Nested
	class TestStatusUnavailableException {

		private final StatusRuntimeException unavailableException = GrpcExceptionTestFactory.createGrpcUnavailableStatusException();

		@Test
		void shouldHaveMessageCode() {
			var response = handleUnavailableException();

			assertThat(response.getBody().getIssues()).hasSize(1);
			assertThat(response.getBody().getIssues().get(0).getMessageCode())
					.isEqualTo(GrpcExceptionController.GRPC_INTERNAL_ERROR_CODE_KEY.getErrorCode());
		}

		@Test
		void shouldHaveMessage() {
			var response = handleUnavailableException();

			assertThat(response.getBody().getIssues()).hasSize(1);
			assertThat(response.getBody().getIssues().get(0).getMessage()).isEqualTo("UNAVAILABLE");
		}

		@Test
		void shouldHaveHttpStatusCode() {
			var response = handleUnavailableException();

			assertThat(response.getStatusCode()).isEqualTo(HttpStatus.SERVICE_UNAVAILABLE);
		}

		@Test
		void shouldNotHaveExceptionId() {
			var response = handleUnavailableException();

			assertThat(response.getBody().getIssues()).hasSize(1);
			assertThat(response.getBody().getIssues().get(0).getExceptionId()).isNotNull();
		}

		private ResponseEntity<ApiError> handleUnavailableException() {
			return handleException(unavailableException);
		}

		@Nested
		class TestWithMetadata {

			private final StatusRuntimeException unavailableExceptionWithMetadata = GrpcExceptionTestFactory
					.createGrpcUnavailableStatusException(GrpcExceptionTestFactory.createMetaData());

			@Test
			void shouldNotHaveExceptionId() {
				var response = handleUnavailableException();

				assertThat(response.getBody().getIssues()).hasSize(1);
				assertThat(response.getBody().getIssues().get(0).getExceptionId()).isEqualTo(GrpcExceptionTestFactory.METADATA_EXCEPTION_ID_VALUE);
			}

			private ResponseEntity<ApiError> handleUnavailableException() {
				return handleException(unavailableExceptionWithMetadata);
			}
		}
	}

	@Nested
	class TestPermissionDeniedException {

		private final StatusRuntimeException permissionDeniedException = GrpcExceptionTestFactory.createGrpcPermissionDeniedStatusException();

		@Test
		void shouldHaveMessageCode() {
			var response = handlePermissionDeniedException();

			assertThat(response.getBody().getIssues()).hasSize(1);
			assertThat(response.getBody().getIssues().get(0).getMessageCode())
					.isEqualTo(GrpcExceptionController.GRPC_PERMISSION_DENIED_CODE_KEY.getErrorCode());
		}

		@Test
		void shouldHaveMessage() {
			var response = handlePermissionDeniedException();

			assertThat(response.getBody().getIssues()).hasSize(1);
			assertThat(response.getBody().getIssues().get(0).getMessage())
					.isEqualTo("PERMISSION_DENIED: " + GrpcExceptionTestFactory.PERMISSION_DENIED_DESCRIPTION);
		}

		@Test
		void shouldHaveHttpStatusCode() {
			var response = handlePermissionDeniedException();

			assertThat(response.getStatusCode()).isEqualTo(HttpStatus.FORBIDDEN);
		}

		@Test
		void shouldNotHaveExceptionId() {
			var response = handlePermissionDeniedException();

			assertThat(response.getBody().getIssues()).hasSize(1);
			assertThat(response.getBody().getIssues().get(0).getExceptionId()).isNotNull();
		}

		private ResponseEntity<ApiError> handlePermissionDeniedException() {
			return handleException(permissionDeniedException);
		}
	}

	private ResponseEntity<ApiError> handleException(StatusRuntimeException exception) {
		return grpcExceptionController.handleStatusRuntimeException(exception);
	}

	@Nested
	class TestMapToIssueParams {

		@Test
		void shouldMapParams() {
			var params = grpcExceptionController.mapToIssueParams(
					GrpcExceptionTestFactory.createMetaDataWithSingleEntry("param_key", "param_value"));

			assertThat(params).hasSize(1);
			assertThat(params.get(0).getName()).isEqualTo("param_key");
			assertThat(params.get(0).getValue()).isEqualTo("param_value");
		}

		@Test
		void shouldNotMapErrorCode() {
			var metadata = GrpcExceptionTestFactory.createMetaDataWithSingleEntry(
					GrpcExceptionController.KEY_ERROR_CODE, GrpcExceptionTestFactory.METADATA_ERROR_CODE_VALUE);

			var params = grpcExceptionController.mapToIssueParams(metadata);

			assertThat(params).isEmpty();
		}

		@Test
		void shouldNotMapExceptionId() {
			var metadata = GrpcExceptionTestFactory.createMetaDataWithSingleEntry(
					GrpcExceptionController.KEY_EXCEPTION_ID, GrpcExceptionTestFactory.METADATA_EXCEPTION_ID_VALUE);

			var params = grpcExceptionController.mapToIssueParams(metadata);

			assertThat(params).isEmpty();
		}
	}

	@DisplayName("Get message code")
	@Nested
	class TestGetMessageCode {

		@Test
		void shouldReturnErrorCode() {
			var messageCodeValue = "ErrorCode";

			var messageCode = getMessageCode(
					new StatusRuntimeException(Status.UNKNOWN, createMetadata(GrpcExceptionController.KEY_ERROR_CODE, messageCodeValue)));

			assertThat(messageCode).isEqualTo(messageCodeValue);
		}

		@Test
		void shouldReturnServiceCode() {
			var messageCodeValue = "ServiceCode";

			var messageCode = getMessageCode(new StatusRuntimeException(Status.UNKNOWN, createMetadata(GrpcUtil.SERVICE_KEY, messageCodeValue)));

			assertThat(messageCode).isEqualTo(messageCodeValue);
		}

		@Test
		void shouldReturnDefault() {
			var messageCode = getMessageCode(new StatusRuntimeException(Status.UNKNOWN));

			assertThat(messageCode).isEqualTo(GrpcExceptionController.GRPC_INTERNAL_ERROR_CODE_KEY.getErrorCode());
		}

		private String getMessageCode(StatusRuntimeException exception) {
			return grpcExceptionController.getMessageCode(exception);
		}

		private Metadata createMetadata(String key, String value) {
			var metadata = new Metadata();
			metadata.put(GrpcUtil.keyOfString(key), value);
			return metadata;
		}
	}
}