package de.ozgcloud.alfa.vorgang;

import static java.util.Objects.*;

import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.springframework.stereotype.Component;

import de.ozgcloud.alfa.common.DateConverter;
import de.xoev.xdomea.Antragsdaten;
import de.xoev.xdomea.AntragsdatenGroupType;
import de.xoev.xdomea.AntragsdatenItemType;
import de.xoev.xdomea.AntragsdatenMultiValueFieldType;
import de.xoev.xdomea.AntragsdatenSingleValueFieldType;
import de.xoev.xdomea.DatatypeType;
import lombok.RequiredArgsConstructor;

@RequiredArgsConstructor
@Component
class FormDataMapper {

	private final DatatypeMapper datatypeMapper;

	public Antragsdaten toAntragsdaten(VorgangWithEingang vorgang) {
		var antragsdaten = new Antragsdaten();
		antragsdaten.getItem().addAll(mapFormData(vorgang));
		antragsdaten.getItem().addAll(mapOtherData(vorgang));
		return antragsdaten;
	}

	List<AntragsdatenItemType> mapFormData(VorgangWithEingang vorgang) {
		return Optional.ofNullable(vorgang.getEingang()).map(Eingang::getFormData).map(this::mapData).orElse(List.of());
	}

	List<AntragsdatenItemType> mapOtherData(VorgangWithEingang vorgang) {
		return Optional.ofNullable(vorgang.getEingang()).map(Eingang::getAntragsteller).map(Antragsteller::getOtherData).map(this::mapData)
				.orElse(List.of());
	}

	List<AntragsdatenItemType> mapData(Map<String, Object> data) {
		return buildItems(WrappedValueExtractor.extractWrappedValues(data));
	}

	List<AntragsdatenItemType> buildItems(Map<?, ?> formData) {
		var items = new ArrayList<AntragsdatenItemType>();
		for (var formDataEntry : formData.entrySet()) {
			var key = (String) formDataEntry.getKey();
			var value = formDataEntry.getValue();
			if (value instanceof Map<?, ?> map) {
				items.add(buildAntragsdatenGroup(key, map));
			} else if (value instanceof Collection<?> collection) {
				items.add(buildMultiValueField(key, collection));
			} else {
				items.add(buildSingleValueField(key, value));

			}
		}
		return items;
	}

	AntragsdatenItemType buildAntragsdatenGroup(String name, Map<?, ?> values) {
		var antragsGroup = new AntragsdatenGroupType();
		antragsGroup.setName(name);
		antragsGroup.getItem().addAll(buildItems(values));
		return antragsGroup;
	}

	AntragsdatenItemType buildMultiValueField(String name, Collection<?> values) {
		var multiValue = new AntragsdatenMultiValueFieldType();
		multiValue.setName(name);
		var datatype = datatypeMapper.from(values);
		multiValue.setType(datatype);
		var multiValueList = multiValue.getValue();
		values.stream().map(v -> formatValue(datatype, v)).forEach(multiValueList::add);
		return multiValue;
	}

	AntragsdatenItemType buildSingleValueField(String name, Object value) {
		var singleValue = new AntragsdatenSingleValueFieldType();
		singleValue.setName(name);
		if (nonNull(value)) {
			var datatype = datatypeMapper.from(value);
			singleValue.setType(datatype);
			singleValue.setValue(formatValue(datatype, value));
		}
		return singleValue;
	}

	Object formatValue(DatatypeType datatype, Object value) {
		return switch (datatype) {
			case STRING -> String.valueOf(value);
			case DATETIME -> DateConverter.toXmlGregorianCalendar((ZonedDateTime) value);
			default -> value;
		};
	}
}
