package de.ozgcloud.alfa.bescheid;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Spy;

import de.ozgcloud.alfa.bescheid.BescheidExportData.BescheidExportDataBuilder;
import de.ozgcloud.alfa.common.binaryfile.BinaryFileService;
import de.ozgcloud.alfa.common.binaryfile.FileId;
import de.ozgcloud.alfa.common.file.OzgFile;
import de.ozgcloud.alfa.common.file.OzgFileTestFactory;
import de.ozgcloud.alfa.common.user.UserProfile;
import de.ozgcloud.alfa.common.user.UserProfileTestFactory;
import de.ozgcloud.alfa.common.user.UserService;
import de.ozgcloud.alfa.export.DokumentTypeTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;
import de.xoev.xdomea.DokumentType;

class ExportBescheidServiceTest {

	@InjectMocks
	@Spy
	private ExportBescheidService service;

	@Mock
	private BescheidService bescheidService;

	@Mock
	private BinaryFileService binaryFileService;

	@Mock
	private DocumentService documentService;

	@Mock
	private UserService userService;

	@Nested
	class TestCreateExportData {
		private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();
		private final Bescheid bescheid = BescheidTestFactory.create();
		private final BescheidExportData exportData = BescheidExportDataTestFactory.create();
		private final BescheidExportInput input = BescheidExportInputTestFactory.create();

		private MockedStatic<BescheidExportData> bescheidExportDataMockedStatic;
		@Mock
		private BescheidExportDataBuilder exportDataBuilder;

		@BeforeEach
		void setUpMocks() {
			bescheidExportDataMockedStatic = mockStatic(BescheidExportData.class);
			bescheidExportDataMockedStatic.when(BescheidExportData::builder).thenReturn(exportDataBuilder);

			doReturn(Stream.of(bescheid)).when(service).getBescheids(vorgang.getId());
			when(exportDataBuilder.build()).thenReturn(exportData);
			doReturn(input).when(service).createBescheidExportInput(bescheid, vorgang);
			doNothing().when(service).addBescheidExportData(input, exportDataBuilder);
		}

		@AfterEach
		void tearDown() {
			bescheidExportDataMockedStatic.close();
		}

		@Test
		void shouldGetBescheids() {
			callService();

			verify(service).getBescheids(vorgang.getId());
		}

		@Test
		void shouldCreateInputData() {
			callService();

			verify(service).createBescheidExportInput(bescheid, vorgang);
		}

		@Test
		void shouldAddInputToBuilder() {
			callService();

			verify(service).addBescheidExportData(input, exportDataBuilder);
		}

		@Test
		void shouldBuildExportData() {
			callService();

			verify(exportDataBuilder).build();
		}

		@Test
		void shouldReturnBuiltExportData() {
			var result = callService();

			assertThat(result).isEqualTo(exportData);
		}

		private BescheidExportData callService() {
			return service.createExportData(vorgang);
		}
	}

	@Nested
	class TestGetBescheids {
		private final String vorgangId = VorgangHeaderTestFactory.ID;
		private final Bescheid bescheid = BescheidTestFactory.createBuilder().status(BescheidStatus.SENT).build();
		private final Bescheid bescheidDraft = BescheidTestFactory.createBuilder().status(BescheidStatus.DRAFT).build();

		@Test
		void shouldCallBescheidService() {
			callService();

			verify(bescheidService).findByVorgangIdSorted(vorgangId);
		}

		@Test
		void shouldReturnBescheids() {
			when(bescheidService.findByVorgangIdSorted(vorgangId)).thenReturn(Stream.of(bescheid));

			var resultBescheids = callService();

			assertThat(resultBescheids).containsExactly(bescheid);
		}

		@Test
		void shouldNotReturnDrafts() {
			when(bescheidService.findByVorgangIdSorted(vorgangId)).thenReturn(Stream.of(bescheid, bescheidDraft));

			var resultBescheids = callService();

			assertThat(resultBescheids).containsExactly(bescheid);
		}

		private Stream<Bescheid> callService() {
			return service.getBescheids(vorgangId);
		}
	}

	@Nested
	class TestCreateBescheidExportInput {
		private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();
		private final Bescheid bescheid = BescheidTestFactory.create();
		private final OzgFile document = OzgFileTestFactory.createWithUniqueId();
		private final OzgFile attachment = OzgFileTestFactory.createWithUniqueId();

		@BeforeEach
		void setUpMocks() {
			doReturn(Stream.of(attachment)).when(service).getAttachments(bescheid);
			doReturn(document).when(service).getDocument(bescheid);
		}

		@Test
		void shouldCallGetAttachents() {
			callService();

			verify(service).getAttachments(bescheid);
		}

		@Test
		void shouldCallGetDocument() {
			callService();

			verify(service).getDocument(bescheid);
		}

		@Test
		void shouldReturnInputData() {
			var resultInput = callService();

			assertThat(resultInput).usingRecursiveComparison()
					.isEqualTo(BescheidExportInputTestFactory.createBuilder().files(List.of(document, attachment)).build());
		}

		private BescheidExportInput callService() {
			return service.createBescheidExportInput(bescheid, vorgang);
		}
	}

	@Nested
	class TestGetDocument {
		private final Bescheid bescheid = BescheidTestFactory.create();
		private final Document document = DocumentTestFactory.create();
		private final OzgFile ozgDocument = OzgFileTestFactory.create();

		@BeforeEach
		void setUpMock() {
			when(documentService.getDocument(bescheid.getBescheidDocument())).thenReturn(document);
		}

		@Test
		void shouldCallDocumentService() {
			callService();

			verify(documentService).getDocument(bescheid.getBescheidDocument());
		}

		@Test
		void shouldCallBinaryFileService() {
			callService();

			verify(binaryFileService).getFile(DocumentTestFactory.FILE_ID);
		}

		@Test
		void shouldReturnOzgFile() {
			when(binaryFileService.getFile(DocumentTestFactory.FILE_ID)).thenReturn(ozgDocument);

			var resultDocument = callService();

			assertThat(resultDocument).isEqualTo(ozgDocument);
		}

		private OzgFile callService() {
			return service.getDocument(bescheid);
		}
	}

	@Nested
	class TestGetAttachments {
		private final Bescheid bescheid = BescheidTestFactory.create();
		private final List<FileId> fileIds = BescheidTestFactory.ATTACHMENTS;

		@Test
		void shouldCallBinaryFileServiceGetFile() {
			service.getAttachments(bescheid);

			verify(binaryFileService).getFiles(fileIds);
		}

		@Test
		void shouldReturnGottenOzgFiles() {
			var expectedOzgFile = OzgFileTestFactory.createWithUniqueId();
			when(binaryFileService.getFiles(fileIds)).thenReturn(Stream.of(expectedOzgFile));

			var resultOzgFiles = service.getAttachments(bescheid);

			assertThat(resultOzgFiles).containsExactly(expectedOzgFile);
		}
	}

	@Nested
	class TestAddBescheidExportData {
		private final List<OzgFile> files = BescheidExportInputTestFactory.FILES;
		private final DokumentType dokumentType = DokumentTypeTestFactory.create();
		private final BescheidExportInput exportInput = BescheidExportInputTestFactory.create();

		@Mock
		private BescheidExportDataBuilder exportDataBuilder;

		@BeforeEach
		void setUpMocks() {
			doReturn(dokumentType).when(service).buildDokumentType(exportInput);
			when(exportDataBuilder.dokumentType(dokumentType)).thenReturn(exportDataBuilder);
		}

		@Test
		void shouldAddAttachmentsToBuilder() {
			callService();

			verify(exportDataBuilder).files(files);
		}

		@Test
		void shouldBuildDokumentType() {
			callService();

			verify(service).buildDokumentType(exportInput);
		}

		@Test
		void shouldAddDokumentTypeToBuilder() {
			callService();

			verify(exportDataBuilder).dokumentType(dokumentType);
		}

		private void callService() {
			service.addBescheidExportData(exportInput, exportDataBuilder);
		}
	}

	@Nested
	class TestBuildDokumentType {
		private final BescheidExportInput exportInput = BescheidExportInputTestFactory.create();
		private final Bescheid bescheid = BescheidExportInputTestFactory.BESCHEID;
		private final List<OzgFile> files = BescheidExportInputTestFactory.FILES;
		private final String fullName = UserProfileTestFactory.FULLNAME;

		private MockedStatic<DokumentTypeBuilder> dokumentTypeBuilderMockedStatic;
		@Mock
		private DokumentTypeBuilder dokumentTypeBuilder;

		@BeforeEach
		void setUp() {
			dokumentTypeBuilderMockedStatic = mockStatic(DokumentTypeBuilder.class);
			dokumentTypeBuilderMockedStatic.when(DokumentTypeBuilder::builder).thenReturn(dokumentTypeBuilder);

			when(dokumentTypeBuilder.withBescheid(bescheid)).thenReturn(dokumentTypeBuilder);
			when(dokumentTypeBuilder.withFiles(files)).thenReturn(dokumentTypeBuilder);
			when(dokumentTypeBuilder.withFullName(fullName)).thenReturn(dokumentTypeBuilder);
			when(dokumentTypeBuilder.withOrganisationseinheitenId(BescheidExportInputTestFactory.ORGANISATIONSEINHEITEN_ID))
					.thenReturn(dokumentTypeBuilder);

			doReturn(fullName).when(service).getFullName(bescheid);
		}

		@AfterEach
		void tearDown() {
			dokumentTypeBuilderMockedStatic.close();
		}

		@Test
		void shouldCallBuilder() {
			callService();

			dokumentTypeBuilderMockedStatic.verify(DokumentTypeBuilder::builder);
		}

		@Test
		void shouldBuildWithBescheid() {
			callService();

			verify(dokumentTypeBuilder).withBescheid(bescheid);
		}

		@Test
		void shouldBuildWithOzgFiles() {
			callService();

			verify(dokumentTypeBuilder).withFiles(files);
		}

		@Test
		void shouldGetFullName() {
			callService();

			verify(service).getFullName(bescheid);
		}

		@Test
		void shouldBuildWithFullName() {
			callService();

			verify(dokumentTypeBuilder).withFullName(fullName);
		}

		@Test
		void shouldBuildWithOrganisationseinheitenId() {
			callService();

			verify(dokumentTypeBuilder).withOrganisationseinheitenId(BescheidExportInputTestFactory.ORGANISATIONSEINHEITEN_ID);
		}

		@Test
		void shouldBuild() {
			callService();

			verify(dokumentTypeBuilder).build();
		}

		@Test
		void shouldReturnBuiltDokumentType() {
			var expectedDokumentType = DokumentTypeTestFactory.create();
			when(dokumentTypeBuilder.build()).thenReturn(expectedDokumentType);

			var resultDokumentType = callService();

			assertThat(resultDokumentType).isEqualTo(expectedDokumentType);
		}

		private DokumentType callService() {
			return service.buildDokumentType(exportInput);
		}
	}

	@Nested
	class TestGetFullName {
		private final Bescheid bescheid = BescheidTestFactory.create();
		private final UserProfile userProfile = UserProfileTestFactory.create();
		private final String expectedFullName = UserProfileTestFactory.FULLNAME;

		@BeforeEach
		void mockUserService() {
			when(userService.getById(SentInfoTestFactory.SENT_BY)).thenReturn(userProfile);
		}

		@Test
		void shouldCallUserServiceGetbyId() {
			callService();

			verify(userService).getById(SentInfoTestFactory.SENT_BY);
		}

		@Test
		void shouldReturnOptionalOfUserProfile() {
			var fullName = callService();

			assertThat(fullName).isEqualTo(expectedFullName);
		}

		private String callService() {
			return service.getFullName(bescheid);
		}
	}
}
